// Code for the PicoPilot v1.1 model
// Board: LilyPad Arduino w/ATmega 328 3.3V
// 
// SPECIFICATIONS 
// This program does the following: 
// - The PicoRover goes forward or backwards gradually when it receives an F or a B from the bluetooth.
// - The PicoRover turns right or left gradually when it receives an R or an L from the bluetooth.
// - An S is received from the bluetooth if the operator does not press any key.
// - An H is received from the bluetooth to switch off the PicoRover, so it can be manipulated.
// - If an order is received and the opposite is sent by the bluetooth, it goes faster until it reaches the commanded area.
// - If it loses coverage, after a second it switches to OFF.

// HOW THE CODE WORKS
// Code operation: The Arduino listens to the serial port, where the bluetooth is connected. If there is something to listen, it
// reads it. If it receives an S, it switches all movement Flags to false and the Attach Flag to true. If it receives an F, a B,
// an R or an L, it activates the Flags of the corresponding movements and it makes sure that the Attach Flag is activated. If
// it receives an H, it switches absolutely all flags to false. Having made these operations, it increases the pointer for the
// received data and it updates the connection time (t3). If there is nothing to listen, every 100ms it processes the information:
// It looks which Flags are activated and which not. If they are activated, it changes the values of dir and pos, that are those
// that it will write to the servos, as appropiate. If the maximum pitch or roll angles are reached, it sets the position to Stop
// or Center to avoid rolling. If the Attach Flag is deactivated, it detaches the servos, so the PicoRover can not move. Finally
// it looks how much time has passed since the last time it received a character. If it has passed more than one second, it will
// disable the Attach Flag, so the servos are locked.


// ESPECIFICACIONES
// Este programa hace lo siguiente:
// - Avanza o retrocede de forma progresiva cuando recibe una F o una B por el bluetooth.
// - Gira de forma progresiva cuando recibe una R o una L por el bluetooth.
// - Se recibe una S por el bluetooth si el operador no pulsa ninguna tecla.
// - Se recibe una H para pasar a DESACTIVADO y poder montar o desmontar el PicoRover.
// - Si ha recibido una orden y se manda la opuesta por el bluetooth, va mas rapido hasta que llegue a la zona ordenada.
// - Si pierde cobertura, pasado un segundo pasa a DESACTIVADO.
//
// COMO FUNCIONA EL CODIGO
// Funcionamiento del codigo: El Arduino escucha por el puerto serie, donde tenemos conectado el BT. Si hay algo para escuchar,
// lo leemos. Si recibimos una S ponemos todos los Flags de movimiento a false y el Flag de Attach a true. Si recibimos una F,
// una B, una R o una L, activamos los Flags de movimientos correspondientes y nos aseguramos que el Flag de Attach este en activo.
// Si recibimos una H, ponemos absolutamente todos los flags a false. Una vez hechas estas operaciones aumentamos el puntero del
// vector de datos recibidos y actualizamos el tiempo de conexion (t3). Si no hay nada para escuchar, cada 100ms pasamos a procesar
// la informacion: miramos que Flags estan activados y cuales no. Si estan activados, variamos los valores de dir y de pos, que son
// los que escribiremos en los servos, segun corresponda. Si el angulo maximo de inclinacion se han superado, colocamos la posicion
// en Stop o Center para evitar que el PicoRover de vueltas. Si el Flag de Attach esta desactivado, haremos unos detach a los
// servos para que el PicoRover no realize ningun movimiento. Por ultimo miramos cuanto tiempo a pasado desde la ultima vez que
// recibimos un caracter. Si ha pasado mas de un segundo, desactivaremos el Flag de Attach para que los servos esten bloqueados.


#define DEBUG // Remove this line to disable online configuration and free some memory

#include <Servo.h>
#ifdef DEBUG
#include <EEPROM.h>
#endif

// Defines used for the servos and the accelerometers
#define ACCELEROMETERX_PIN 0 // Analog input pin number to read the X axis acceleration
#define ACCELEROMETERY_PIN 1 // Analog input pin number to read the Y axis acceleration
#define ACCELEROMETERZ_PIN 2 // Analog input pin number to read the Z axis acceleration
#define THRUST_PIN 9      // Ouput pin number to thrust the PicoRover 5V
#define STEERING_PIN 10   // Ouput pin number to steer the PicoRover 5V

#define BUFFER_SIZE 512   // Input buffer size in bytes

// Setup parameters                                                    Cmd  EEPROM-Address
#define CENTER 130         // Steering neutral position in degrees      A     0
#define STRLEFT 180        // Steering left limit in degrees            B     4
#define STRRIGHT 80        // Steering right limit in degrees           C     8
#define STOP 71            // Thrust neutral position in degrees        D    12
#define THRFWD 75          // Thrust forward limit in degrees           E    16
#define THRBCK 66          // Thrust backward limit in degrees          F    20
#define THRSTEP 0.5        // Slow thrust step in degrees               G    24
#define THRSTEPHI 1        // Fast thrust step in degrees               H    28
#define STRSTEP 5          // Slow steering step in degrees             I    32
#define STRSTEPHI 10       // Fast steering step in degrees             J    36
#define SERVO_TIME 100     // Servo update interval in milliseconds     K    40
#define STANDBY_TIME 1000  // Stand by delay time in milliseconds       L    44

// Parameters
float Center;
float StrLeft;
float StrRight;
float Stop;
float ThrFwd;
float ThrBck;
float ThrStep;
float ThrStepHi;
float StrStep;
float StrStepHi;
int ServoTime;
int StandbyTime;

// Global variables
char data[BUFFER_SIZE];
int i;
unsigned long t0;  // CPU initial time in milliseconds
unsigned long t1;  // CPU previous time in milliseconds
unsigned long dt;  // Current CPU interval in milliseconds
unsigned long t2;
unsigned long t3;
Servo thrust;      // Create servo object to control a servo
Servo steering;    // Create servo object to control a servo
float dir;         // Current thrust value
float pos;         // Current steering value
float lastdir;     // Last thrust value
float lastpos;     // Last steering value
int AccelX = 0;    // Accelerometer X axis value
int AccelY = 0;    // Accelerometer Y axis value
int AccelZ = 0;    // Accelerometer Z axis value

// Control flags
boolean FlagForward = false;
boolean FlagBackward = false;
boolean FlagRight = false;
boolean FlagLeft = false;
boolean FlagAttach;
boolean LastAttach;
boolean FlagPos = false;   // Send time and position to the serial port
boolean FlagAccel = false; // Send accelerometers data to the serial port

#ifdef DEBUG

// Read a float variable from EEPROM address
float EEPROMGetFloat(int addr)
{
  float num;
  ((byte*)&num)[0] = EEPROM.read(addr++);
  ((byte*)&num)[1] = EEPROM.read(addr++);
  ((byte*)&num)[2] = EEPROM.read(addr++);
  ((byte*)&num)[3] = EEPROM.read(addr++);
  //Serial.print("Read "); Serial.print(addr-4); Serial.print("="); Serial.println(num); delay(100);
  return num;
}

// Write a float variable to EEPROM address
void EEPROMSetFloat(int addr, float num)
{
  //Serial.print("Write "); Serial.print(addr); Serial.print("="); Serial.println(num); delay(100);
  EEPROM.write(addr++, ((byte*)&num)[0]);
  EEPROM.write(addr++, ((byte*)&num)[1]);
  EEPROM.write(addr++, ((byte*)&num)[2]);
  EEPROM.write(addr++, ((byte*)&num)[3]);
}

// Read a character from serial port
byte SerialGetByte()
{
  byte n = 0;
  unsigned long t = millis();
  do
  {
    if(Serial.available()>0)
    {
      n = Serial.read();
      break;
    }
  }
  while(millis()-t < 100);
  return n;
}

// Read a float variable from serial port
//  (with no exponent and terminated with a non numeric character)
float SerialGetFloat()
{
  byte n;
  float num = 0;
  int comma = 1;
  unsigned long t = millis();
  do
  {
    if(Serial.available()>0)
    {
      n = Serial.read();
      if(n=='.')
      {
        comma = 10;
      }
      else if(n>='0' && n<='9')
      {
        if(comma<=1)
        {
          num = num * 10 + (float)(n - '0');
        }
        else
        {
          num += ((float)(n - '0')) / ((float)comma);
          comma *= 10;
        }
      }
      else break;
    }
  }
  while(millis()-t < 100);
  //Serial.print("SerialRead="); Serial.println(num); delay(100);
  return num;
}

// Load setup parameters from EEPROM
void LoadParameters()
{
  int n = 0;
  Center = EEPROMGetFloat(n); n+=4;
  StrLeft = EEPROMGetFloat(n); n+=4;
  StrRight = EEPROMGetFloat(n); n+=4;
  Stop = EEPROMGetFloat(n); n+=4;
  ThrFwd = EEPROMGetFloat(n); n+=4;
  ThrBck = EEPROMGetFloat(n); n+=4;
  ThrStep = EEPROMGetFloat(n); n+=4;
  ThrStepHi = EEPROMGetFloat(n); n+=4;
  StrStep = EEPROMGetFloat(n); n+=4;
  StrStepHi = EEPROMGetFloat(n); n+=4;
  ServoTime = (int)EEPROMGetFloat(n); n+=4;
  StandbyTime = (int)EEPROMGetFloat(n); n+=4;
}

// Save setup parameters to EEPROM
void SaveParameters()
{
  int n = 0;
  EEPROMSetFloat(n, Center); n+=4;
  EEPROMSetFloat(n, StrLeft); n+=4;
  EEPROMSetFloat(n, StrRight); n+=4;
  EEPROMSetFloat(n, Stop); n+=4;
  EEPROMSetFloat(n, ThrFwd); n+=4;
  EEPROMSetFloat(n, ThrBck); n+=4;
  EEPROMSetFloat(n, ThrStep); n+=4;
  EEPROMSetFloat(n, ThrStepHi); n+=4;
  EEPROMSetFloat(n, StrStep); n+=4;
  EEPROMSetFloat(n, StrStepHi); n+=4;
  EEPROMSetFloat(n, (float)ServoTime); n+=4;
  EEPROMSetFloat(n, (float)StandbyTime); n+=4;
}

#endif

// Initialize parameters
void InitParameters()
{
  Center = CENTER;
  StrLeft = STRLEFT;
  StrRight = STRRIGHT;
  Stop = STOP;
  ThrFwd = THRFWD;
  ThrBck = THRBCK;
  ThrStep = THRSTEP;
  ThrStepHi = THRSTEPHI;
  StrStep = STRSTEP;
  StrStepHi = STRSTEPHI;
  ServoTime = SERVO_TIME;
  StandbyTime = STANDBY_TIME;
}

// Initialization of the PicoRover
void setup()
{
  InitParameters();
#ifdef DEBUG
  //SaveParameters(); // Use this line to initialize EEPROM with default parameters
  LoadParameters();
#endif
  Serial.begin(4800);
  t0 = millis();
  t3 = millis();
  dir = Stop;
  pos = Center;
  lastdir = -1;
  lastpos = -1;
  FlagAttach = false;
  LastAttach = FlagAttach;
  i = 0;
}

//Loop start for the PicoRover movements
void loop()
{
  int n;
  
  t2 = millis();
  
  // Read commands from serial port if available
  if(Serial.available()>0)
  {
    t3 = t2;
    LastAttach = FlagAttach;
    data[i] = Serial.read();
    switch(data[i])
    {
      
      // The PicoRover must stop
      case 'S':
        FlagForward = false;
        FlagBackward = false;
        FlagRight = false;
        FlagLeft = false;
        FlagAttach = true;
        break;
      
      // The PicoRover must go forward
      case 'F':
        FlagForward = true;
        FlagBackward = false;
        FlagAttach = true;
        if(i!=0 && data[i-1]=='F')
        {
          FlagRight = false;
          FlagLeft = false;
        }
        break;
      
      // The PicoRover must go backwards
      case 'B':
        FlagBackward = true;
        FlagForward = false;
        FlagAttach = true;
        if(i!=0 && data[i-1]=='B')
        {
          FlagRight = false;
          FlagLeft = false;
        }
        break;
      
      // The PicoRover must turn right
      case 'R':
        FlagRight = true;
        FlagLeft = false;
        FlagAttach = true;
        if(i!=0 && data[i-1]=='R')
        {
          FlagForward = false;
          FlagBackward = false;
        }
        break;
      
      // The PicoRover must turn left
      case 'L':
        FlagLeft = true;
        FlagRight = false;
        FlagAttach = true;
        if(i!=0 && data[i-1]=='L')
        {
          FlagForward = false;
          FlagBackward = false;
        }
        break;
      
      // The PicoRover must halt
      case 'H':
        FlagAttach = false;
        pos = Center;
        dir = Stop;
        //__asm__ __volatile__ ( "sleep" "\n\t" :: );
        break;
      
#ifdef DEBUG
      // A configuration parameter must be changed
      case 'C':
        n = SerialGetByte();
        if(n>='A' && n<='L')
        {
          EEPROMSetFloat((n-'A')<<2, SerialGetFloat());
          LoadParameters();
        }
        else 
        {
          SerialGetByte(); // Read ';'
          // Reset to default parameters
          if(n=='0')
          {
            InitParameters();
            SaveParameters();
          }
          // Show current parameters
          else if(n=='1')
          {
            for(n='A'; n<='L'; n++)
            {
            Serial.print('C'); Serial.print(n, BYTE);
            Serial.print(EEPROMGetFloat((n-'A')<<2));
            Serial.println(";");
            }
          }
          // Show time and position
          else if(n=='2')
          {
            FlagPos = !FlagPos;
          }
          // Show inertial sensor
          else if(n=='3')
          {
            FlagAccel = !FlagAccel;
          }
        }
        break;
#endif
      
      // If any of the previous cases happens
      default:
        break;
    }
    i++;
  }
  else
  {
    // Anti-roll system: it reads the vertical accelerometer Y
    // and if the value is negative, it blocks the PicoRover Thrust or Steering
    AccelX = analogRead(ACCELEROMETERX_PIN)-512;
    AccelY = analogRead(ACCELEROMETERY_PIN)-512;
    //AccelZ = analogRead(ACCELEROMETERZ_PIN)-512;
    AccelZ = 0;
#ifdef DEBUG
    if(FlagAccel)
    {
      Serial.print("A,");
      Serial.print(AccelX); Serial.print(",");
      Serial.print(AccelY); Serial.print(",");
      Serial.print(AccelZ); Serial.print(";");
      delay(100);
    }
#endif
    
    // Check servo interval
    t1=millis();
    if(t1-t0 > ServoTime)
    {
      dt = t1-t0;
      t0 = t1;
      lastdir = dir;
      lastpos = pos;
      
      // Attach and initialize servos when the first command is received
      if(LastAttach==false && FlagAttach==true)
      {
        LastAttach = true;
        thrust.attach(THRUST_PIN);
        steering.attach(STEERING_PIN);
        thrust.write((int)dir);
        steering.write((int)pos);
        delay(100);
      }
      
      // Forward is selected
      if(FlagForward)
      {
        // Anti-roll check
        if(AccelY<0)
        {
          dir = Stop;
        }
        else if(dir>Stop) // Select slow or fast step
        {
          dir += ThrStep;
        }
        else
        {
          dir += ThrStepHi;
        }
      }
      
      // Backward is selected
      if(FlagBackward)
      {
        // Anti-roll check
        if(AccelY<0)
        {
          dir = Stop;
        }
        else if(dir<Stop) // Select slow or fast step
        {
          dir -= ThrStep;
        }
        else
        {
          dir -= ThrStepHi;
        }
      }
      
      // Right is selected
      if(FlagRight)
      {
        // Anti-roll check
        if(AccelY<0)
        {
          pos = Center;
        }
        else if(pos<Center) // Select slow or fast step
        {
          pos -= StrStep;
        }
        else
        {
          pos -= StrStepHi;
        }
      }
      
      // Left is selected
      if(FlagLeft)
      {
        // Anti-roll check
        if(AccelY<0)
        {
          pos = Center;
        }
        else if(pos>Center) // Select slow or fast step
        {
          pos += StrStep;
        }
        else
        {
          pos += StrStepHi;
        }
      }
      
      // Center if no thrust is selected
      if(FlagForward==false && FlagBackward==false)
      {
        if(dir > Stop+ThrStep)
        {
          dir -= ThrStep;
        }
        else if(dir < Stop-ThrStep)
        {
          dir += ThrStep;
        }
        else // Dead zone
        {
          dir = Stop;
        }
      }
      
      // Center if no steering is selected
      if(FlagRight==false && FlagLeft==false)
      {
        if(pos > Center+StrStep)
        {
          pos -= StrStep;
        }
        else if(pos < Center-StrStep)
        {
          pos += StrStep;
        }
        else // Dead zone
        {
          pos = Center;
        }
      }
      
      // Check steering limits
      if(pos>StrLeft)
      {
        pos = StrLeft;
      }
      else if(pos<StrRight)
      {
        pos = StrRight;
      }
      
      //Write steering position to servo
      if(FlagAttach && pos!=lastpos)
      {
        if(steering.attached()==false)
        {
          steering.attach(STEERING_PIN); // Attach if needed
        }
        steering.write((int)pos);
      }
      else if(steering.attached()) // Detach if not needed
      {
        steering.detach();
      }
      
      // Check thrust limits
      if(dir>ThrFwd)
      {
        dir = ThrFwd;
      }
      else if(dir<ThrBck)
      {
        dir = ThrBck;
      }
      
      // Write thrust position to servo
      if(FlagAttach && dir!=Stop)
      {
        if(thrust.attached()==false)
        {
          thrust.attach(THRUST_PIN); // Attach if needed
        }
        thrust.write((int)dir);
      }
      else if(thrust.attached())
      {
        thrust.detach(); // Detach if not needed
      }
    }

    // Check standby time and detach servos
    if (t2-t3 > StandbyTime)
    {
      thrust.detach();
      steering.detach();
      FlagAttach = false;
      FlagForward = false;
      FlagBackward = false;
      FlagRight = false;
      FlagLeft = false;
    }
    //Serial.print(dir); Serial.print(","); Serial.println(pos);
  }
  
  // Check input buffer overrun
  if(i==BUFFER_SIZE)
  {
    i = 0;
    // Serial.println("The buffer is full. Proceeding to clean up...");
  }
}

